# -*- coding: utf-8 -*-
import re, urlparse
from httplib import HTTPConnection, _CS_IDLE


def raw_pipeline(domain, pages, max_out_bound=4, method='GET', timeout=None, debuglevel=0):
    pagecount = len(pages)
    conn = HTTPConnection(domain, timeout=timeout)
    conn.set_debuglevel(debuglevel)
    respobjs = [None] * pagecount
    finished = [False] * pagecount
    responses = [None] * pagecount
    data = [None] * pagecount
    headers = {'Host': domain, 'Content-Length': 0, 'Connection': 'Keep-Alive'}

    while not all(finished):
        # Send
        out_bound = 0
        for i, page in enumerate(pages):
            if out_bound >= max_out_bound:
                break
            elif page and not finished[i] and respobjs[i] is None:
                if debuglevel > 0:
                    print 'Sending request for %r...' % (page,)
                conn._HTTPConnection__state = _CS_IDLE # FU private variable!
                conn.request(method, page, None, headers)
                respobjs[i] = conn.response_class(conn.sock, strict=conn.strict, method=conn._method)
                out_bound += 1

        # Try to read a response
        for i, resp in enumerate(respobjs):
            if resp is None:
                continue
            if debuglevel > 0:
                print 'Retrieving %r...' % (pages[i],)
            out_bound -= 1
            skip_read = False
            resp.begin()
            if debuglevel > 0:
                print '    %d %s' % (resp.status, resp.reason)
            if 200 <= resp.status < 300:
                # Ok
                data[i] = resp.read()
                cookie = resp.getheader('Set-Cookie')
                if cookie is not None:
                    headers['Cookie'] = cookie
                skip_read = True
                finished[i] = True
                responses[i] = respobjs[i]
                respobjs[i] = None
            elif 300 <= resp.status < 400:
                # Redirect
                loc = resp.getheader('Location')
                respobjs[i] = None
                parsed = loc and urlparse.urlparse(loc)
                if not parsed:
                    # Missing or empty location header
                    data[i] = (resp.status, resp.reason)
                    finished[i] = True
                elif parsed.netloc != '' and parsed.netloc != domain:
                    # Redirect to another domain
                    data[i] = (resp.status, resp.reason, loc)
                    finished[i] = True
                else:
                    path = urlparse.urlunparse(parsed._replace(scheme='',netloc='',fragment=''))
                    if debuglevel > 0:
                        print '  Updated %r to %r' % (pages[i],path)
                    pages[i] = path
            elif resp.status >= 400:
                # Failed
                data[i] = (resp.status, resp.reason)
                finished[i] = True
                responses[i] = respobjs[i]
                respobjs[i] = None
            if resp.will_close:
                # Connection (will be) closed, need to resend
                conn.close()
                if debuglevel > 0:
                    print '  Connection closed'
                for j, f in enumerate(finished):
                    if not f and respobjs[j] is not None:
                        if debuglevel > 0:
                            print '  Discarding out-bound request for %r' % (pages[j],)
                        respobjs[j] = None
                break
            elif not skip_read:
                resp.read() # read any data
            if any(not f and respobjs[j] is None for j,f in enumerate(finished)):
                # Send another pending request
                break
        else:
            break # All respobjs are None?
    return responses, data


def pipeline(urls, **kwargs):
    url_index = dict((u, i) for i, u in enumerate(urls))
    responses, data = [None] * len(urls), [''] * len(urls)

    # Grouping urls by domain
    urls = sorted(re.match(r'http://([^/]+)(.*)$',u).groups() for u in url_index)
    groups = group(urls, 0)

    # Executing requests in each group
    for g in groups:
        pages = [url[1] for url in g]
        g_responses, g_data = raw_pipeline(g.grouper, pages, **kwargs)

        for i, page in enumerate(pages):
            url = 'http://%s%s' % (g.grouper, page)
            index = url_index[url]
            responses[index] = g_responses[i]
            data[index] = data[i]

    # Some urls may have been doubled,
    # copy responses and data for them
    if len(url_index) < len(urls):
        for i, res in enumerate(responses):
            if res is None:
                index = url_index[urls[i]]
                responses[i] = responses[index]
                data[i] = data[index]

    return responses, data


def group(object_list, key, min_group=1):
    groups = []
    group = None
    prev = None

    for obj in object_list:
        val = getattr(obj, key) if isinstance(key, str) and hasattr(obj, key) else obj[key]
        if val != prev:
            if group and len(group) < min_group:
                group.grouper += ', ' + val
            else:
                group = Group()
                group.grouper = val
                groups.append(group)
        group.append(obj)
        prev = val

    return groups


if __name__ == '__main__':
    domain = 'en.wikipedia.org:80'
    pages = ('/wiki/HTTP_pipelining', '/wiki/HTTP', '/wiki/HTTP_persistent_connection')
    data = raw_pipeline(domain, pages, debuglevel=1, method='HEAD', max_out_bound=6, timeout=1)
    for i,page in enumerate(data):
        print
        print '==== Page %r ====' % (pages[i],)
        #print page[:512]



# -*- coding: utf-8 -*-
from functools import wraps

from django.db import transaction, connections


def commit_on_success_on(*databases):
    databases = set(databases)
    def decorator(func):
        for db in databases:
            func = transaction.commit_on_success(db)(func)
        return wraps(func)(func)
    return decorator


def queryset_iterator(queryset, chunksize=1000):
    pk = 0
    try:
        last_pk = queryset.order_by('-pk')[0].pk
    except IndexError:
        raise StopIteration
    queryset = queryset.order_by('pk')
    while pk < last_pk:
        # Use iterator to bypass queryset cache
        for row in queryset.filter(pk__gt=pk)[:chunksize].iterator():
            pk = row.pk
            yield row

def queryset_chunks(queryset, chunksize=1000):
    pk = 0
    while True:
        chunk = list(queryset.order_by('pk').filter(pk__gt=pk)[:chunksize])
        if chunk:
            pk = chunk[-1].pk
            yield chunk
        else:
            break


### A couple of low-level utilities

def fetch_all(sql, params=(), server='default'):
    return do_sql(sql, params, server).fetchall()

def fetch_row(sql, params=(), server='default'):
    return first(fetch_all(sql, params, server))

def fetch_col(sql, params=(), server='default'):
    return [row[0] for row in fetch_all(sql, params, server)]

def fetch_val(sql, params=(), server='default'):
    return first(fetch_row(sql, params, server))

def do_sql(sql, params=(), server='default'):
    cursor = connections[server].cursor()
    cursor.execute(sql, params)
    return cursor


def first(seq):
    return next(seq, None)



# -*- coding: utf-8 -*-
from django import forms
from django.utils.safestring import mark_safe


class SimpleWidget(forms.Widget):
    def __init__(self, format='%(value)s'):
        super(SimpleWidget, self).__init__()
        self.format = format

    def render(self, name, value, attrs=None):
        return mark_safe(self.format % {'name': name, 'value': value})


class CommaSeparatedInput(forms.TextInput):
    def _format_value(self, value):
        if isinstance(value, basestring):
            return value
        else:
            return ','.join(map(unicode, value))

    def value_from_datadict(self, data, files, name):
        value = data.get(name, '').strip()
        if value in ['', None]:
            return []
        return value.split(',')


class MultilineInput(forms.Textarea):
    def _format_value(self, value):
        if isinstance(value, basestring):
            return value
        else:
            return '\n'.join(value)



# -*- coding: utf-8 -*-
from itertools import chain, izip, tee


def with_prev(coll, key=None):
    if key is not None:
        coll = sorted(coll, key=key)
    a, b = tee(coll)
    return izip(a, chain([None], b))

def with_next(coll, key=None):
    if key is not None:
        coll = sorted(coll, key=key)
    a, b = tee(coll)
    return izip(a, padnone(skip(b)))

def padnone(iterable):
    """Returns the sequence elements and then returns None indefinitely.

    Useful for emulating the behavior of the built-in map() function.
    """
    return chain(iterable, repeat(None))



# -*- coding: utf-8 -*-
import re, string, random

from django.utils.safestring import mark_safe


def random_str(length=8, letters=string.ascii_letters + string.digits):
    """Generates random string of latin characters and digits"""
    return "".join(random.choice(letters) for x in range(length))


def formatnumber(value, splitter=' '):
    s = unicode(value)
    l = len(s)
    indexes = range(-(l + -l % 3), 0, 3)
    return mark_safe(splitter.join(s[i:i+3 or None] for i in indexes))


def formatnumber_html(value, splitter='<small>&nbsp;</small>'):
    return mark_safe(formatnumber(value, splitter))



# -*- coding: utf-8 -*-
from functools import wraps
from datetime import datetime

from django.conf import settings
from django.template import loader, RequestContext
from django.http import HttpResponse
from django.utils import simplejson as json
from django.utils.http import http_date


def template_guess(func):
    template_dir = func.__module__.replace('.views', '').replace('.', '/')
    return template_dir + '/' +                                                     \
           func.__name__ + getattr(settings, 'TEMPLATE_DEFAULT_EXTENSION', '.html')


def render_dict(request, output):
    # extracting hook keys
    template = output.pop('TEMPLATE')
    kwargs = dict((k.lower(), output.pop(k)) for k in ('STATUS', 'CONTENT_TYPE') if k in output)
    # rendering remplate and wrapping it into HttpResponse
    rendered = loader.render_to_string(template, output, RequestContext(request))
    return HttpResponse(rendered, **kwargs)


def render_to(template=None):
    """
    Renders view result to template. Inspired by `@render_to()` decorator
    from django-annoying package.
    """
    def decorator(func):
        template_default = template or template_guess(func)

        @wraps(func)
        def wrapper(request, *args, **kwargs):
            output = func(request, *args, **kwargs)
            if not isinstance(output, dict):
                return output

            output.setdefault('TEMPLATE', template_default)
            response = render_dict(request, output)
            if 'COOKIES' in output:
                cookies = (output['COOKIES'],) if isinstance(output['COOKIES'], dict) else output['COOKIES']
                for i in cookies:
                    response.set_cookie(**i)
            return response

        return wrapper
    return decorator


def _json_default(obj):
    # TODO: unify json func here, in ajax, and the one used by JSONField
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError

def render_to_json(ensure_ascii=True, default=_json_default):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            response = func(*args, **kwargs)
            if isinstance(response, HttpResponse):
                return response
            elif isinstance(response, str):
                json_data = response
            else:
                if settings.DEBUG:
                    json_data = json.dumps(response, sort_keys=True, indent=4, ensure_ascii=False, default=default)
                else:
                    json_data = json.dumps(response, ensure_ascii=ensure_ascii, separators=(',',':'), default=default)
            return HttpResponse(json_data, mimetype='application/json; charset=utf-8')
        return wrapper
    return decorator


def last_modified(func):
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        response = func(request, *args, **kwargs)
        response['Last-Modified'] = http_date()
        return response
    return wrapper



# -*- coding: utf-8 -*-
import re, sys, traceback
from smtplib import SMTPRecipientsRefused

from django.conf import settings
from django.core.mail import EmailMessage, mail_admins as django_mail_admins
from django.template import loader, RequestContext
from django.core.files.uploadedfile import InMemoryUploadedFile


def parse_email_data(content):
    lines = content.splitlines()

    subject, content_subtype = None, None
    headers = {}

    for i, line in enumerate(lines):
        # Headers read?
        if not line and subject:
            lines = lines[i+1:]
            break

        m = re.search(r'^([\w-]+): (.+)', line.strip())
        if not m:
            raise ValueError('Email subject required\nPrepend email content with "Subject:\s<subject>\\n\\n"')
        name, value = m.groups()

        if name == 'Subject':
            subject = value
        elif name == 'Content-Type':
            m = re.search('^text/([\w-]+)', value)
            if not m:
                raise ValueError('Erroneous email content type %s' % value)
            content_subtype = m.group(1)
        else:
            headers[name] = value
    return subject, content_subtype, headers, '\n'.join(lines)


def render_to_email(email, template_name, data=None, request=None, from_email=None, attachment=None):
    if not email:
        return

    context = RequestContext(request) if request else None
    content = loader.render_to_string(template_name, data or {}, context)

    subject, content_subtype, headers, email_text = parse_email_data(content)

    if not isinstance(email, (list, tuple)):
        email = [email]

    message = EmailMessage(subject, email_text, from_email, email, headers=headers)
    if content_subtype:
        message.content_subtype = content_subtype

    if attachment:
        add_attach = lambda uploadedFile: message.attach(uploadedFile.name, uploadedFile.read(), uploadedFile.content_type)

        if isinstance(attachment, list):
            for attach in attachment:
                if isinstance(attach, InMemoryUploadedFile):
                    add_attach(attach)
        elif isinstance(attachment, InMemoryUploadedFile):
            add_attach(attachment)

    try:
        message.send()
    except SMTPRecipientsRefused:
        pass

    if settings.DEBUG:
        print u'To: %s' % ', '.join(message.to)
        print u'From: %s' % message.from_email
        print u'Subject: %s' % message.subject
        print message.body


def mail_admins(subject, message='', trace=True):
    if trace:
        exc_info = sys.exc_info()
        if exc_info:
            message += ("\n\n" if message else '') + ''.join(traceback.format_exception(*exc_info))
    django_mail_admins(subject, message, fail_silently=True)



# -*- coding: utf-8 -*-
import re, datetime
from functools import wraps
import pytz

from django.http import HttpResponse
try:
    from django.utils.timezone import get_default_timezone
except ImportError:
    from django.conf import settings
    _default_timezone = pytz.timezone(settings.TIME_ZONE)
    get_default_timezone = lambda: _default_timezone

from .decorators import render_to_json


class AjaxException(Exception):
    def __init__(self, error='', **kwargs):
        defaults = {'success': False, 'error': error, 'data': {}}
        self.params = dict(defaults, **kwargs)

class Ajax(object):
    error = AjaxException

    def __call__(self, func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            try:
                data = func(request, *args, **kwargs)
                if isinstance(data, HttpResponse):
                    return data
                return {'success': True, 'data': data}
            except self.error, e:
                return e.params
        return render_to_json(default=encode_object)(wrapper)

    def catch(self, *exceptions):
        assert exceptions, 'at least one exception type is required'
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except self.error:
                    raise
                except exceptions, e:
                    error_name = camel_to_underscores(e.__class__.__name__)
                    error_text = e.args[0] if e.args else ''
                    raise self.error(error_name, error_text=error_text)
            return wrapper
        return decorator

    def login_required(self, func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            if request.user.is_anonymous():
                raise self.error('login_required')
            else:
                return func(request, *args, **kwargs)
        return wrapper

ajax = Ajax()


def camel_to_underscores(name):
    words = re.findall(r'[A-Z][a-z0-9]+', name)
    return '_'.join(w.lower() for w in words)


### Encoding object to json
from pytz.tzinfo import DstTzInfo
DstTzInfo.__json__ = lambda self: self.zone

UTC = pytz.timezone('UTC')
EPOCH = datetime.datetime(1970,1,1,0,0,0, tzinfo=UTC)

# TODO: add support for querysets and models
def encode_object(obj):
    if hasattr(obj, '__json__'):
        return obj.__json__()
    elif isinstance(obj, datetime.datetime):
        delta = to_default(obj) - EPOCH
        return {'_type': 'date', 'ms': int(delta.total_seconds() * 1000)}
    elif isinstance(obj, datetime.date):
        return datetime.datetime(*obj.timetuple()[:6])
    else:
        raise TypeError('%s is not JSON serializable' % obj.__class__.__name__)

def to_default(dt):
    if dt.tzinfo is None:
        return get_default_timezone().localize(dt)
    else:
        return dt.astimezone(get_default_timezone())




# -*- coding: utf-8 -*-
try:
    import cPickle as pickle
except ImportError:
    import pickle

import re, datetime
from django.db import models
from django import forms
from django.core import exceptions, validators
from django.utils.text import capfirst
from django.utils import simplejson as json
from django.utils.translation import ugettext_lazy as _
from django.conf import settings

from handy.forms import CommaSeparatedInput, MultilineInput
from handy.utils import get_module_attr


__all__ = ['AdditionalAutoField', 'AdditionalAutoFieldManager',
           'BigAutoField', 'IntegerArrayField', 'BigIntegerArrayField',
           'StringArrayField', 'BigIntegerField', 'JSONField']


class AdditionalAutoField(models.Field):
    """
    Дополнительный AutoField делает ещё одно auto поле не являющееся primary key
    Не поддерживается всеми бэкендами! viva la PostgreSQL :)

    Требуется использовать AdditionalAutoFieldManager в качестве базового менеджера, чтобы null не пропускались в базу:
        class MyModel(models.Model):
            _base_manager = AdditionalAutoFieldManager()
    """
    empty_strings_allowed = False
    def __init__(self, *args, **kwargs):
        kwargs['blank'] = True
        models.Field.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "AutoField"

    def to_python(self, value):
        if value is None:
            return value
        try:
            return int(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(
                _("This value must be an integer."))

    def get_db_prep_value(self, value):
        if value is None:
            return None
        return int(value)

    def contribute_to_class(self, cls, name):
        models.Field.contribute_to_class(self, cls, name)

    def formfield(self, **kwargs):
        if self.primary_key:
            return None
        defaults = {'form_class': forms.IntegerField, 'widget': forms.HiddenInput}
        defaults.update(kwargs)
        return models.Field.formfield(self, **defaults)


class AdditionalAutoFieldManager(models.Manager):
    def _insert(self, values, **kwargs):
        # Значение по умолчанию выставлется базой
        values = [v for v in values if v[1] is not None or not isinstance(v[0], AdditionalAutoField)]
        return super(AdditionalAutoFieldManager, self)._insert(values, **kwargs)


class BigAutoField(models.AutoField):
    def __init__(self, verbose_name=None, name=None, external_sequence=False, **kwargs):
        self.external_sequence = external_sequence
        models.AutoField.__init__(self, verbose_name, name, **kwargs)

    def db_type(self, connection):
        return 'bigint' if self.external_sequence else 'bigserial'

    # Использует патч ForeignKey
    def rel_db_type(self, connection):
        return 'bigint'


class UntypedMultipleField(forms.Field):
    def __init__(self, *args, **kwargs):
        self.coerce = kwargs.pop('coerce', lambda val: val)
        super(UntypedMultipleField, self).__init__(*args, **kwargs)


class TypedMultipleField(UntypedMultipleField):
    def to_python(self, value):
        value = super(TypedMultipleField, self).to_python(value)
        if value not in validators.EMPTY_VALUES:
            try:
                value = map(self.coerce, value)
            except (ValueError, TypeError):
                raise exceptions.ValidationError(self.error_messages['invalid'])
        return value

class TypedMultipleChoiceField(TypedMultipleField, forms.MultipleChoiceField):
    def validate(self, value):
        pass


class ArrayField(models.Field):
    def validate(self, value, model_instance):
        # Это по-большей части скопировано с Field.validate
        # К сожалению приходится копипастить т.к. там неправильно проверяется
        # принадлежность к choices для многозначных полей
        if not self.editable:
            return

        if value is None and not self.null:
            raise exceptions.ValidationError(self.error_messages['null'])

        if not self.blank and value in validators.EMPTY_VALUES:
            raise exceptions.ValidationError(self.error_messages['blank'])

        # TODO: validate choices

    def formfield(self, **kwargs):
        if self.choices:
            # Тоже скопипащено из Field.formfield
            defaults = {
                'choices': self.choices,
                'coerce': self.coerce,
                'required': not self.blank,
                'label': capfirst(self.verbose_name),
                'help_text': self.help_text,
                'widget': forms.CheckboxSelectMultiple
            }
            defaults.update(kwargs)
            return TypedMultipleChoiceField(**defaults)
        else:
            defaults = {
                'form_class': TypedMultipleField,
                'coerce': self.coerce,
                'widget': CommaSeparatedInput,
            }
            defaults.update(kwargs)
            return super(ArrayField, self).formfield(**defaults)


class IntegerArrayField(ArrayField):
    default_error_messages = {
        'invalid': _(u'Enter only digits separated by commas.')
    }
    description = _("Array of integers")
    coerce = int

    def db_type(self, connection):
        return 'int[]'


class BigIntegerArrayField(IntegerArrayField):
    def db_type(self, connection):
        return 'bigint[]'

# Fix unicode arrays for postgresql_psycopg2 backend
try:
    import psycopg2
    psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)
except ImportError:
    pass

class StringArrayField(ArrayField):
    description = _("Array of strings")
    coerce = unicode

    def __init__(self, verbose_name=None, **kwargs):
        kwargs.setdefault('max_length', 127)
        super(StringArrayField, self).__init__(verbose_name, **kwargs)

    def db_type(self, connection):
        return 'varchar(%s)[]' % self.max_length


# Функции для кодирования/декодирования объектов в json
def default_decode((cls, data)):
    cls = get_module_attr(cls)
    obj = cls.__new__(cls)
    obj.__dict__.update(data['__data__'])
    return obj

def encode_object(obj):
    if hasattr(obj, '__json__'):
        data = obj.__json__()
        if isinstance(data, tuple):
            decode, data = data
        else:
            decode = default_decode
            data = ('%s.%s' % (obj.__class__.__module__, obj.__class__.__name__), data)

        if callable(decode):
            decode = '%s.%s' % (decode.__module__, decode.__name__)

        return {
            '__decode__': decode,
            '__data__': data
        }
    else:
        return dict(__pickled__=pickle.dumps(obj))

def decode_object(d):
    if '__decode__' in d:
        return get_module_attr(d['__decode__'])(d['__data__'])
    elif '__pickled__' in d:
        return pickle.loads(str(d['__pickled__']))
    else:
        return d

class JSONField(models.TextField):
    """JSONField is a generic textfield that neatly serializes/unserializes
    JSON objects seamlessly"""
    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        """Convert our string value to JSON after we load it from the DB"""
        if value == "":
            return None

        try:
            if isinstance(value, basestring):
                return json.loads(value, object_hook=decode_object)
        except ValueError:
            pass

        return value

    def get_prep_value(self, value):
        """Convert our JSON object to a string before we save"""
        if value == "" or value is None:
            return None

        #if isinstance(value, dict):
        value = json.dumps(value, default=encode_object, ensure_ascii=False, separators=(',',':'))

        return super(JSONField, self).get_prep_value(value)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': forms.Field,
        }
        defaults.update(kwargs)
        defaults['widget'] = JSONTextarea
        return super(JSONField, self).formfield(**defaults)


class JSONTextarea(forms.Textarea):
    def value_from_datadict(self, data, files, name):
        value = data.get(name, '').strip()
        if value in ['', None]:
            return {}
        return json.loads(value)

    def render(self, name, value, attrs=None):
        return super(JSONTextarea, self).render(name, json.dumps(value), attrs=attrs)



class BigIntegerField(models.IntegerField):
    def db_type(self):
        # только для mysql и postgres
        return "bigint"

    def get_internal_type(self):
        return "BigIntegerField"

    def to_python(self, value):
        if value is None:
            return value
        try:
            return long(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(
                _("This value must be a long integer."))


# Прописываем правила для South, пустые так как новых, неунаследованных атрибутов у нас нет
try:
    from south.modelsinspector import add_introspection_rules
    rules = [
        ((BigAutoField, ), [], {"external_sequence": ["external_sequence", {"default": False}]}),
        ((StringArrayField, ), [], {"max_length": ["max_length", {}]}),
    ]
    add_introspection_rules(rules, ["^handy\.models\.fields\."])
except ImportError:
    pass
